// ------ Libraries and Definitions ------
#include "simpletools.h"
#include "ws2812.h"
#include "fdserial.h"
#include "ping.h"


//--------- Music data -----------------------------
// Chromatic scale in sharps
// s stands for sharp
const int  C  = 0;
const int  Cs = 1;
const int  D  = 2;
const int  Ds = 3;
const int  E  = 4;
const int  F  = 5;
const int  Fs = 6;
const int  G  = 7;
const int  Gs = 8;
const int  A  = 9;
const int  As = 10;
const int  B  = 11;
// The enharmonic spellings for the sharps
// b stands for flat
const int Db  = 1;
const int Eb  = 3;
const int Gb  = 6;
const int Ab  = 8;
const int Bb  = 10;

const char* chromaticSharps[] = {"C","C#","D","D#","E","F","F#","G","G#","A","A#","B"};
//-------------------------------------------------


//--------- Pins -----------
// Define our Propeller pins

// The pitch LASERPing
#define PITCH_PIN       16

// The volume LASERPing
#define VOLUME_PIN      17

// The pitch LEDs
#define PITCH_LEDS_PIN  25

// The buttons
#define SCALE_BUTTON_PIN   3
#define NOTE_BUTTON_PIN    2
#define OCTAVE_BUTTON_PIN  1
//-------------------------


//------ Misc constants ----
#define NUM_LEDS        48
#define NUM_PITCHES     24
#define NUM_OCTAVES     10
#define NUM_SCALES      9
#define X               255 // Not a note
//--------------------------


// -------- Colors -----------------------------
// You can find some predefined ones in ws2812.h
#define PITCH_LED_COLOR   0x00994b
#define VOLUME_LED_COLOR  0xff00ff
#define DARK_GREEN        0x001100
#define DARK_RED          0x110000
#define DARK_YELLOW       0x111100
// ---------------------------------------------


//--------- Scales -----------------------------
typedef struct Scale{
  char name[20];
  int root;
  int *offsets;
  int *hilights;
} Scale;  

Scale currentScale;
Scale scales[NUM_SCALES];
//---------------------------------------------


//------- MIDI interface ------
// Serial interface for MIDI
fdserial *midi;
//-----------------------------


//------- Serial LCD interface ------
serial *lcd;
const int ON  = 22;
const int CLR = 12;
const int LINE1 = 128;
const int LINE2 = 148;
const int LINE3 = 168;
const int BACKLIGHT_ON = 17;
const int BACKLIGHT_OFF = 18;
//----------------------------------


//---------- ws2812 LEDs --------------------
ws2812 *leds;
// Array of RGB values for each LED
int  ledsArray[] = {0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0};
//-------------------------------------------


//--------- Pitch information -------------------------
int newPitch     = 0;
int currentPitch = 0;
int pitchLED     = 0;
int midLED       = 0; // A pitch between two MIDI notes
//-----------------------------------------------------


//--------- Volume information --------
int newVolume = 0;
int currentVolume = 0;
int volumeLED = 0;
//-------------------------------------


//----------------- Scales ---------------------------------------------------------
// A scale defined as a sequence of MIDI number offsets from a root.
// We have 24 slots for pitches, because we have 24 LEDs.
// Every other LED corresponds to a note. The other LEDs show speaces between notes.
// Not every slot has has to have a note in it. 
// X is equal to 255, and stands for not a note.
// Hilight convention:
// Root  = DARK_RED
// Third = DARK_YELLOW
// Fifth = DARK_GREEN

int ionianOffsets[] = {0,X,2,X,4,X,5,X,7,X,9,X,11,X,12,X,14,X,16,X,17,X,19,X};
int ionianHilights[] = {DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0};
  
int aeolianOffsets[] = {0,X,2,X,3,X,5,X,7,X,8,X,10,X,12,X,14,X,15,X,17,X,19,X};
int aeolianHilights[] = {DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0};
                                                      
int mixolydianOffsets[] = {0,X,2,X,4,X,5,X,7,X,9,X,10,X,12,X,14,X,16,X,17,X,19,X};
int mixolydianHilights[] = {DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0};
                                                       
int chromaticOffsets[] = {0,X,1,X,2,X,3,X,4,X,5,X,6,X,7,X,8,X,9,X,10,X,11,X};
int chromaticHilights[] = {DARK_RED,0,0,0,0,0,0,0,DARK_YELLOW,0,0,0,0,0,DARK_GREEN,0,0,0,0,0,0,0,0,0};

int majorPentatonicOffsets[] = {0,X,2,X,4,X,7,X,9,X,12,X,14,X,16,X,19,X,21,X,24,X,26,X};
int majorPentatonicHilights[] = {DARK_RED,0,0,0,DARK_YELLOW,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0,DARK_YELLOW,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0};   
                                  
int minorPentatonicOffsets[] = {0,X,3,X,5,X,7,X,10,X,12,X,15,X,17,X,19,X,22,X,24,X,27,X};
int minorPentatonicHilights[] = {DARK_RED,0,0,0,0,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0,0,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0};     

int bluesOffsets[] = {0,X,3,X,5,X,6,X,7,X,10,X,12,X,15,X,17,X,18,X,19,X,22,X};
int bluesHilights[] = {DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0,0,0};       
  
int spanishPhyrgianOffsets[] = {0,X,1,X,4,X,5,X,7,X,8,X,10,X,12,X,13,X,16,X,17,X,19,X};
int spanishPhyrgianHilights[] = {DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0};   

int flamencoOffsets[] = {0,X,2,X,3,X,6,X,7,X,8,X,11,X,12,X,14,X,15,X,18,X,19,X};
int flamencoHilights[] = {DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0};    
//----------------------------------------------------------------------------------       


//----------------- LASERVox ergonomics --------------------------------------------
// Adjust these values to set the range of action of the pitch and volume sensors.

// Min and max distances for the pitch LASERPing.
// This is the range in microseconds over which the pitch LASERPing operates.
int pitchMinDist = 600;  // Distance of the closest note to the sensor
int pitchMaxDist = 3500; // Distance of the furthest note from the sensor
int pitchOffDist = 5000; // Distance for auto off feature


// Min and max distances for the volume LASERPing.
// This is the range in microseconds over which the volume LASERPing operates.
int volumeMinDist = 600;  // Distance of volume 0
int volumeMaxDist = 2000; // Distance of volume 127
//----------------------------------------------------------------------------------



//---- Refresh flags -------------------
// 1 = refresh needed
// 0 = no refresh needed
//volatile int refreshLEDs = 0;
volatile int refreshTuning = 0; 
//--------------------------------------


//--------- The current scale ---------- 
volatile int currentScaleIndex = 0;
volatile int currentRoot = 60;
volatile int currentNote = 0;  
volatile int currentOctave = 4;
//-------------------------------------- 



int MIDILock;
int LEDLock;
//int LCDLock;



// ------ Function Declarations ------
//------------- Cores ---------------- 
void getPitch();
void getVolume();
void getButtons();  
//------------------------------------

// LEDs
void updateLEDs();

// LCD
void printScaleToLCD();

// Tuning
void updateTuning();

// Utility
int MIDINum(int, int);

//------------- MIDI -----------------
// MIDI messages
void noteOn(int note, int volume);
void noteOff(int note);
void channelVolume(int value);
void allNotesOff(void);
//------------------------------------


// ------ Main Program ------
int main() {
  // Print a message to the console
  print("HELLO WORLD %d", MIDINum(C, 4));
  
  // Set up the LCD
  lcd = fdserial_open(24, 24, 0, 2400);
  writeChar(lcd, ON);
  writeChar(lcd, CLR);
  writeChar(lcd, BACKLIGHT_ON);
  pause(5);
  dprint(lcd, "     LASERVox");
  writeChar(lcd, LINE2);
  dprint(lcd, "Jim Arlow");
  writeChar(lcd, LINE3);
  dprint(lcd, "Nuts&Volts, 2021");
  pause(2000);
  
  
  
  // Set up some locks. 
  MIDILock = locknew();
  LEDLock = locknew();
  //LCDLock = locknew();

  

  //------- Set up the scales ---------

  // Ionian
  strcpy(scales[0].name, "Ionian");
  scales[0].root = MIDINum(C, 4);
  scales[0].offsets = ionianOffsets;
  scales[0].hilights = ionianHilights;
  
  // Aeolian
  strcpy(scales[1].name, "Aeolian");
  scales[1].root = MIDINum(C, 4);
  scales[1].offsets = aeolianOffsets;
  scales[1].hilights = aeolianHilights;

  // Mixolydian  
  strcpy(scales[2].name, "Mixolydian");
  scales[2].root = MIDINum(G, 3);
  scales[2].offsets = mixolydianOffsets;
  scales[2].hilights = mixolydianHilights; 
  
  // Chromatic;
  strcpy(scales[3].name, "Chromatic");
  scales[3].root = MIDINum(C, 4);
  scales[3].offsets = chromaticOffsets;
  scales[3].hilights = chromaticHilights;
  
  // Major Pentatonic;
  strcpy(scales[4].name, "Major Pentatonic");
  scales[4].root = MIDINum(C, 4);
  scales[4].offsets = majorPentatonicOffsets;
  scales[4].hilights = majorPentatonicHilights;
  
  // Minor Pentatonic;
  strcpy(scales[5].name, "Minor Pentatonic");
  scales[5].root = MIDINum(C, 4);
  scales[5].offsets = minorPentatonicOffsets;
  scales[5].hilights = minorPentatonicHilights;

  // Blues
  strcpy(scales[6].name, "Blues");
  scales[6].root = MIDINum(C, 4);
  scales[6].offsets = bluesOffsets;
  scales[6].hilights = bluesHilights;
  
  // Spanish Phrygian
  strcpy(scales[7].name, "Spanish Phrygian");
  scales[7].root = MIDINum(C, 4);
  scales[7].offsets = spanishPhyrgianOffsets;
  scales[7].hilights = spanishPhyrgianHilights;

  // Flamenco
  strcpy(scales[8].name, "Flamenco");
  scales[8].root = MIDINum(C, 4);
  scales[8].offsets = flamencoOffsets;
  scales[8].hilights = flamencoHilights;
  
  // Set the current scale
  currentScale = scales[0];
  updateTuning();
  
  //----------------------------------
  
  // Open MIDI connection
  midi = fdserial_open(-1, 0, 0, 31250);
  
  // Open connection to the 48 LEDs
  leds = ws2812b_open(); 
   
  // Turn the LEDs on
  updateLEDs();
  
  // Turn off any already playing notes
  allNotesOff();
  
  // Launch the volume, pitch and button cogs
  cog_run(getVolume, 128);
  cog_run(getPitch, 128);
  cog_run(getButtons, 128);

  // Main loop
  while(1) {}

}


void updateTuning(){
  currentScale = scales[currentScaleIndex];
  currentScale.root = currentRoot;     
  printScaleToLCD();
}  

// ------ Functions -------------------------------------------------

void updateLEDs() {
  // Lighting the WS2812 is a two phase process:
  // 1) Set the correct RGB values into the array for the LEDs
  // 2) Submit the array to the WS2812 interface to turn on the LEDs
  // Note that nothing happens until we submit the changes!
      
  // Turn off all pitch LEDs that are not hilighted
  for(int i=0; i <= NUM_PITCHES; i++){
    ledsArray[i] = currentScale.hilights[i];
    }
    
  // Turn off all volume LEDs
  for(int i=NUM_PITCHES; i <= NUM_LEDS; i++){
    ledsArray[i] = 0;
    }


  // Illuminate the LEDs between the pitches
  ledsArray[midLED] = COLOR_BROWN;
  // Illuminate the current pitch and volume LEDs
  //ledsArray[pitchLED] = 0x00994b;  
  ledsArray[pitchLED] = PITCH_LED_COLOR;   
  ledsArray[volumeLED] = VOLUME_LED_COLOR;
  
  // Publish the changes to illuminate the LEDs
  lockset(LEDLock);
  ws2812_set(leds, PITCH_LEDS_PIN, ledsArray, NUM_LEDS);
  lockclr(LEDLock);
}


// This function runs in its own core, and continually polls
// the pitch Ping sensor.
void getPitch() {
  int dist = 0;
  int rawDist = 0;
  int scaleIndex = 0;
  int tmp = 0;
  
  while(1) {
    // Measure the distance from the pitch sensor in microseconds
    rawDist = ping(PITCH_PIN);
    
    // Auto off feature:
    // If the player is out of the playing range, turn all notes off
    if (rawDist >= pitchOffDist)
    {
      // Turn off any currently playing note
      allNotesOff();
      // Once we are out of range, whatever the current
      // pitch was, it is now invalid, so set it to 0 so
      // that it is recalculated next time around.
      currentPitch = 0;
      pitchLED = 23;
      // Update the LEDs
      updateLEDs();
      continue;
    }
    
    // Constrain the distance to be between pitchMinDist and pitchMaxDist
    dist = constrainInt(rawDist, pitchMinDist, pitchMaxDist);


    // Calculate the index into the scale
    scaleIndex = NUM_PITCHES - 1 - mapInt(dist, pitchMinDist, pitchMaxDist, 0, NUM_PITCHES - 1);

    // If this is a valid note, then update the newPitch
    tmp = currentScale.offsets[scaleIndex];
    if( tmp != X ){
      newPitch = tmp + currentScale.root;
    }    
     
    // Turn on the LED between pitches
    midLED = scaleIndex;
    
    // Process pitch change.
    if (newPitch != currentPitch) {
      // Turn off any currently playing note
      noteOff(currentPitch);

      // Set the current note to the new note
      currentPitch = newPitch;

      // Turn on the current note at amximum velocity                 
      noteOn(currentPitch, 127);
      // Adjust the volume of the synthesizer
      channelVolume(currentVolume);
      
      // Set the pitch LED
      pitchLED = scaleIndex;
    }
    
    // Update the LEDs
    updateLEDs();
  }
}


// This function runs in its own core, and continually polls
// the volume Ping sensor.
void getVolume() {
  int dist = 0;
  while(1) {
    // Measure the distance from the volume sensor in microseconds
    // Constrain the value to be between volumeMinDist and volumeMaxDist
    dist = constrainInt(ping(VOLUME_PIN), volumeMinDist, volumeMaxDist);
    
    // The max MIDI range for volume is 0 to 127
    newVolume = mapInt(dist,volumeMinDist,volumeMaxDist,0,127);
    
    // Calculate the LED for the current volume
    volumeLED = mapInt(dist,volumeMinDist,volumeMaxDist,24,47);
    
    // Process volume change
    if (newVolume != currentVolume) {
      currentVolume = newVolume;
      // The MIDI Channel Volume message changes the volume
      // whilst a note is playing.
      channelVolume(currentVolume);
      // Update the LEDs
      updateLEDs();
    }

  }
}

// Turn on a note
// Note: 0..127 from C-1 to G8
// Volume: 0..127
void noteOn(int note, int volume){
  lockset(MIDILock);
  fdserial_txChar(midi,144);
  fdserial_txChar(midi,note);
  fdserial_txChar(midi,volume);
  lockclr(MIDILock); 
}  

// Turn off a note
// Note: 0..127 from C-1 to G8
void noteOff(int note){
  lockset(MIDILock);
  fdserial_txChar(midi,128);
  fdserial_txChar(midi,note);
  fdserial_txChar(midi,0);
  lockclr(MIDILock); 
}  

void allNotesOff(){
  for(int i = 0; i <= 127; i++){
    noteOff(i);
  }    
}  
// Change the volume of a note that is playing.
// The volume parameter sent as part of Note On
// determines the initial loudness of the note.
// Channel Volume controls the loadness of an
// note WHILST IT IS SOUNDING. It is essential
// for an eNApressive instrument
void channelVolume(int value){
  lockset(MIDILock); 
  fdserial_txChar(midi,176);
  fdserial_txChar(midi,7);
  fdserial_txChar(midi,value);
  lockclr(MIDILock); 
}  


// Convert a note and an octave to a MIDI number
int MIDINum(int note, int octave){
  return note + octave*12 + 12;
}  


// This function runs in its own core, and continually polls
// the buttons.
void getButtons(){
  // Increment the following circular counters on a positive edge.
  // currentOctave (0..9)
  // currentScale (0..6)
  // currentNote (0..11)
  while(1)                                  
  {
    // The octave button
    if (input(OCTAVE_BUTTON_PIN) > 0){ 
      currentOctave = (++currentOctave) % NUM_OCTAVES;
      currentRoot = currentOctave * 12 +  currentNote;
      updateTuning();
      pause(500);
    }     
    
    // The scale button
    if (input(SCALE_BUTTON_PIN) > 0){ 
      currentScaleIndex = (++currentScaleIndex) % NUM_SCALES;
      updateTuning();
      pause(500);
    }     
    
    // The note button
    if (input(NOTE_BUTTON_PIN) > 0){ 
      currentNote = (++currentNote) % 12;
      currentRoot = currentOctave * 12 +  currentNote;
      updateTuning();
      pause(500);
    }                                                          
  }
}

// Output the current scale to the LCD
void printScaleToLCD(){
  int octave = currentScale.root / 12 - 1;
  int note = currentScale.root % 12;
  int scaleNote;
  // Clear the LCD
  writeChar(lcd, CLR);
  pause(5);
  dprint(lcd, "Root = %s%d", chromaticSharps[note], octave);
  // Move to line 2
  writeChar(lcd, LINE2);
  dprint(lcd, "%s",currentScale.name);
  // Move to line 3
  writeChar(lcd, LINE3);
  for(int i=0; i<=23; i+=2){
    scaleNote = (currentScale.root + currentScale.offsets[i])%12;
    dprint(lcd, "%s ", chromaticSharps[scaleNote]);
  }
}  