// ------ Libraries and Definitions ------
#include "simpletools.h"
#include "ws2812.h"
#include "fdserial.h"
#include "ping.h"

// Chromatic scale in sharps
// s stands for sharp
#define C   0
#define Cs  1
#define D   2
#define Ds  3
#define E   4
#define F   5
#define Fs  6
#define G   7
#define Gs  8
#define A   9
#define Aa  10
#define B   11
// The enharmonic spellings for the sharps
// b stands for flat
#define Db  1
#define Eb  3
#define Gb  6
#define Ab  8
#define Bb  10

// Define our Propeller pins
#define PITCH_PIN 16
#define VOLUME_PIN 17
#define PITCH_LEDS_PIN   25
#define VOLUME_LEDS_PIN   7
#define NUM_LEDS 48
#define NUM_PITCHES 24
#define X 255

// Some colors
// You can find some predefined ones in ws2812.h
#define PITCH_LED_COLOR 0x00994b
#define VOLUME_LED_COLOR 0xff00ff
#define DARK_GREEN 0x001100
#define DARK_RED   0x110000
#define DARK_YELLOW 0x111100

typedef struct Scale{
  int root;
  int *offsets;
  int *hilights;
} Scale;

Scale currentScale;

// ------ Global Variables and Objects ------
// Serial interface for MIDI
fdserial *midi;

// LED interface for ws2812 smart LEDs
ws2812 *leds;
// Array of RGB values for each LED
int  ledsArray[] = {0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0};

// Store the pitch information
int newPitch = 0;
int currentPitch = 0;
int pitchLED = 0;

// Store the volume information
int newVolume = 0;
int currentVolume = 0;
int volumeLED = 0;

// Indicate a pitch between two MIDI notes
int midLED = 0;

// The scale defined as MIDI number offsets from the root.
// We have 24 slots for notes, because we have 24 LEDs, if we had more LEDs, we could
// have more slots.
// Not every slot has has to have a note in it.
// X is equal to 255, and stands for not applicable, or not a note.
// Convention whre possible:
// Root = Red
// Third = Yellow
// Fifth = Green
const int NumScales = 8;

int mixolydianOffsets[] = {0,X,2,X,4,X,5,X,7,X,9,X,10,X,12,X,14,X,16,X,17,X,19,X};
int mixolydianHilights[] = {DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0};

int ionianOffsets[] = {0,X,2,X,4,X,5,X,7,X,9,X,11,X,12,X,14,X,16,X,17,X,19,X};
int ionianHilights[] = {DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0};

int chromaticOffsets[] = {0,X,1,X,2,X,3,X,4,X,5,X,6,X,7,X,8,X,9,X,10,X,11,X};
int chromaticHilights[] = {DARK_RED,0,0,0,0,0,0,0,DARK_YELLOW,0,0,0,0,0,DARK_GREEN,0,0,0,0,0,0,0,0,0};

int majorPentatonicOffsets[] = {0,X,2,X,4,X,7,X,9,X,12,X,14,X,16,X,19,X,21,X,24,X,26,X};
int majorPentatonicHilights[] = {DARK_RED,0,0,0,DARK_YELLOW,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0,DARK_YELLOW,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0};

int minorPentatonicOffsets[] = {0,X,3,X,5,X,7,X,10,X,12,X,15,X,17,X,19,X,22,X,24,X,27,X};
int minorPentatonicHilights[] = {DARK_RED,0,0,0,0,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0,0,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0};

int bluesOffsets[] = {0,X,3,X,5,X,6,X,7,X,10,X,12,X,15,X,17,X,18,X,19,X,22,X};
int bluesHilights[] = {DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0,0,0,DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0,0,0};

int spanishPhyrgianOffsets[] = {0,X,1,X,4,X,5,X,7,X,8,X,10,X,12,X,13,X,16,X,17,X,19,X};
int spanishPhyrgianHilights[] = {DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,DARK_YELLOW,0,0,0,DARK_GREEN,0};

int flamencoOffsets[] = {0,X,2,X,3,X,6,X,7,X,8,X,11,X,12,X,14,X,15,X,18,X,19,X};
int flamencoHilights[] = {DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0,0,0,0,0,DARK_RED,0,0,0,0,0,0,0,DARK_GREEN,0};

// Min and max distances for the pitch LASERPing.
// This is the range over which the pitch LASERPing operates.
int pitchMinDist = 200;  // Distance of the closest note to the sensor
int pitchMaxDist = 3000; // Distance of the furthest note from the sensor
int pitchOffDist = 5000;


// Min and max distances for the volume LASERPing.
// This is the range over which the volume LASERPing operates.
int volumeMinDist = 200;  // Distance of volume 0
int volumeMaxDist = 2000; // Distance of volume 127


// Flag to indicate when LEDs need to be refreshed.
// 1 = refresh needed
// 0 = no refresh needed
volatile int refreshLEDs = 0;


// Buttons ----------------------
// The scale
const int ScaleButtonPin = 3;
volatile int currentScaleIndex = 0;

// The note
const int NoteButtonPin = 2;
const int NumNotes = 12;
volatile int currentNote = 0;

// The octave
const int OctaveButtonPin = 1;
const int NumOctaves = 10;
volatile int currentOctave = 4;


volatile int refreshTuning = 0;
//------------------------------


// ------ Function Declarations ------
// LEDs
void updateLEDs();

// Pitch and volume (these run as separate cores)
void getPitch();
void getVolume();

// Main MIDI messages
void noteOn(int note, int volume);
void noteOff(int note);
void channelVolume(int value);

// Useful MIDI messages
void allNotesOff(void);
void allSoundOff(void);

// Some other MIDI messages
void breath(int value);
void portamentoOn();
void portamentoOff();
void portamentoControl(int val); // The amount of portamento glide
void portamentoTime(int amount); // The speed of the portamento glide

int MIDINum(int, int);

void getButtons();

// ------ Main Program ------
int main() {
  print("HELLO WORLD %d", MIDINum(C, 4));

  Scale scales[NumScales];

  scales[0].root = MIDINum(G, 3);
  scales[0].offsets = mixolydianOffsets;
  scales[0].hilights = mixolydianHilights;

  // Initialize the scales
  Scale mixolydian;
  mixolydian.root = MIDINum(G, 3);
  mixolydian.offsets = mixolydianOffsets;
  mixolydian.hilights = mixolydianHilights;

  Scale ionian;
  ionian.root = MIDINum(C, 4);
  ionian.offsets = ionianOffsets;
  ionian.hilights = ionianHilights;

  Scale chromatic;
  chromatic.root = MIDINum(C, 4);
  chromatic.offsets = chromaticOffsets;
  chromatic.hilights = chromaticHilights;

  Scale majorPentatonic;
  majorPentatonic.root = MIDINum(C, 4);
  majorPentatonic.offsets = majorPentatonicOffsets;
  majorPentatonic.hilights = majorPentatonicHilights;

  Scale minorPentatonic;
  minorPentatonic.root = MIDINum(C, 4);
  minorPentatonic.offsets = minorPentatonicOffsets;
  minorPentatonic.hilights = minorPentatonicHilights;


  // Set the current scale
  //currentScale = mixolydian;
  //currentScale = ionian;
  currentScale = scales[0];
  //currentScale = majorPentatonic;
  //currentScale = minorPentatonic;

  // Open MIDI connection
  midi = fdserial_open(-1, 0, 0, 31250);

  // Open connection to the 48 LEDs
  leds = ws2812b_open();

  // Turn the LEDs on
  updateLEDs();

  // Turn off any already playing notes
  allNotesOff();

  // Launch the pitch and volume cogs
  cog_run(getVolume, 128);
  cog_run(getPitch, 128);
  cog_run(getButtons, 128);

  while(1) {
    // When a pitch change or a volume change is detected
    // by the pitch or volume cores, the refreshLEDs flag
    // is set to 1
    if(refreshLEDs == 1){
      updateLEDs();
      refreshLEDs=0;
    }
  }

}


// ------ Functions ------
void updateLEDs() {
  // Lighting the WS2812 is a two phase process:
  // 1) Set the correct RGB values into the array for the LEDs
  // 2) Submit the array to the WS2812 interface to turn on the LEDs
  // Note that nothing happens until we submit the changes!

  // Turn off all pitch LEDs that are not hilighted
  for(int i=0; i <= NUM_PITCHES; i++){
    ledsArray[i] = currentScale.hilights[i];
    }

  // Turn off all volume LEDs
  for(int i=NUM_PITCHES; i <= NUM_LEDS; i++){
    ledsArray[i] = 0;
    }


  // Illuminate the LEDs between the pitches
  ledsArray[midLED] = COLOR_BROWN;
  // Illuminate the current pitch and volume LEDs
  //ledsArray[pitchLED] = 0x00994b;
  ledsArray[pitchLED] = PITCH_LED_COLOR;
  ledsArray[volumeLED] = VOLUME_LED_COLOR;

  // Publish the changes to illuminate the LEDs
  ws2812_set(leds, PITCH_LEDS_PIN, ledsArray, NUM_LEDS);
}


// This function runs in its own core, and continually polls
// the pitch Ping sensor.
void getPitch() {
  int dist = 0;
  int rawDist = 0;
  int scaleIndex = 0;
  int tmp = 0;
  while(1) {
    // Measure the distance from the pitch sensor in microseconds and
    // constrain it to our defined range
    rawDist = ping(PITCH_PIN);
    dist = constrainInt(rawDist, pitchMinDist, pitchMaxDist);

    // Auto off feature:
    // If the player is out of the playing range, turn all notes off
    if (rawDist >= pitchOffDist)
    {
      allNotesOff();
      pitchLED = 23;
      refreshLEDs = 1;
      continue;
    }

    // Calculate the index into the scale
    scaleIndex = NUM_PITCHES - 1 - mapInt(dist, pitchMinDist, pitchMaxDist, 0, NUM_PITCHES - 1);

    // If this is a valid note, then update the newPitch
    tmp = currentScale.offsets[scaleIndex];
    if( tmp != X ){
      newPitch = tmp + currentScale.root;
    }

    // Turn on the LED between pitches
    midLED = scaleIndex;

    // Process pitch change.
    // It is essential to completely process the pitch change in this core
    // so that nothing can change as the propeller "rotates" through the other cores.
    if (newPitch != currentPitch) {
      // Turn off the previously playing note
      noteOff(currentPitch);

      // Set the current note to the new note
      currentPitch = newPitch;

      // Turn on the current note
      noteOn(currentPitch, 127);
      //channelVolume(currentVolume);

      // Set the pitch LED
      pitchLED = scaleIndex;
    }
    // Always refresh the LEDs
    refreshLEDs=1;
  }
}


// This function runs in its own core, and continually polls
// the volume Ping sensor.
void getVolume() {
  int dist = 0;
  while(1) {
    // Measure the distance from the volume sensor in microseconds
    // Constrain the value to be between volumeMinDist and volumeMaxDist
    dist = constrainInt(ping(VOLUME_PIN), volumeMinDist, volumeMaxDist);

    // The max MIDI range for volume is 0 to 127
    newVolume = mapInt(dist,volumeMinDist,volumeMaxDist,0,127);

    // Calculate the LED for the current volume
    volumeLED = mapInt(dist,volumeMinDist,volumeMaxDist,24,47);

    // Process volume change
    if (newVolume != currentVolume) {
      currentVolume = newVolume;
      // The MIDI Channel Volume message changes the volume
      // whilst a note is playing.
      channelVolume(currentVolume);
      // There has been a volume change, so raise a flag
      // to signal an update is needed for the LEDs
      refreshLEDs=1;
    }

  }
}

// Turn on a note
// Note: 0..127 from C-1 to G8
// Volume: 0..127
void noteOn(int note, int volume){
    fdserial_txChar(midi,144);
    fdserial_txChar(midi,note);
    fdserial_txChar(midi,volume);
}

// Turn off a note
// Note: 0..127 from C-1 to G8
void noteOff(int note){
    fdserial_txChar(midi,128);
    fdserial_txChar(midi,note);
    fdserial_txChar(midi,0);
}

// Change the volume of a note that is playing.
// The volume parameter sent as part of Note On
// determines the initial loudness of the note.
// Channel Volume controls the loadness of an
// note WHILST IT IS SOUNDING. It is essential
// for an eNApressive instrument
void channelVolume(int value){
    fdserial_txChar(midi,176);
    fdserial_txChar(midi,7);
    fdserial_txChar(midi,value);
}

// Here are some Channel Mode messages that
// effet the MIDI channel itself rather than
// individual notes or controllers

// All Notes Off
void allNotesOff(void){
    fdserial_txChar(midi,123);
    fdserial_txChar(midi,0);
}

// All Sound Off
void allSoundOff(void){
    fdserial_txChar(midi,120);
    fdserial_txChar(midi,0);
}


// Here are some CC (control change) messages.

// What these do depends on your synth and selected voice.
// The CC status byte is 176.

// Breath control
void breath(int value){
    fdserial_txChar(midi,176);
    fdserial_txChar(midi,2);
    fdserial_txChar(midi,value);
}


// Portamento on
void portamentoOn(){
    fdserial_txChar(midi,176);
    fdserial_txChar(midi,65);
    fdserial_txChar(midi,127); // >=64
}

// Portamento off
void portamentoOff(){
    fdserial_txChar(midi,176);
    fdserial_txChar(midi,65);
    fdserial_txChar(midi,0);  // <=63
}

// The amount of portamento glide
void portamentoControl(int val){
    fdserial_txChar(midi,176);
    fdserial_txChar(midi,84);
    fdserial_txChar(midi,val);
}

// The speed of the portamento glide
void portamentoTime(int amount){
    fdserial_txChar(midi,176);
    fdserial_txChar(midi,5);
    fdserial_txChar(midi,amount);
}

// Convert a note and an octave to a MIDI number
int MIDINum(int note, int octave){
  return note + octave*12 + 12;
}


void getButtons(){
  // Increment the following circular counters on a positive edge.
  // currentOctave (0..9)
  // currentScale (0..6)
  // currentNote (0..11)
  while(1)
  {
    // The octave button
    if (input(OctaveButtonPin) > 0){
      currentOctave = (++currentOctave) % NumOctaves;
      refreshTuning = 1;
      pause(500);
    }

    // The scale button
    if (input(ScaleButtonPin) > 0){
      currentScaleIndex = (++currentScaleIndex) % NumScales;
      refreshTuning = 1;
      pause(500);
    }

    // The note button
    if (input(NoteButtonPin) > 0){
      currentNote = (++currentNote) % NumNotes;
      refreshTuning = 1;
      pause(500);
    }
  }
}
