/*
  Jim Arlow, 2020
  Program to test MIDI output on pin 0 and LED output on pin 25. 
  Demonstrates use of locks and the cogid() function.
  The Propeller chip has seven available locks. 
  lockset(int lockid) causes a core to wait until the specified lock is free and then sets it. 
  lockclr(int lockid) clears a lock so another core can use it.
  We will use this mechanism to ensure only one cog at a time can access midi and leds.
*/
#include "simpletools.h"  
#include "fdserial.h" 
#include "ws2812.h"  

#define MIDI_OUT_PIN 0  
#define LEDS_PIN 25 
#define NUM_LEDS 48   

// Viva La Musica, Micheal Praetorius (1571-1621)
// This is a three voice round.
int vivaNotes[] = {79,72,71,69,67,66,67,66,71,72,74,72,71,69,71,69,67,67,60,62,64,62};  
int vivaDurations[] = {3, 1, 2, 1, 1, 2, 2, 4, 3, 1, 2, 1, 1, 2, 2, 4, 2, 4, 2, 2, 2, 4}; // 1 = quaver, 2 = crotchet etc.  

const int numNotes = 22;
const int tempo = 200;      // The duration of a quaver in milliseconds
const int twoBarsRest = 16; // The rest between voices

// The MIDI serial output
fdserial *midi; 

// The LED port
ws2812 *leds;  
// Array of RGB values for each LED
int  ledsArray[] = {0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0,
                    0,0,0,0,0,0,0,0,0,0,0,0};


// We will protect the resources with locks
// so that only one core may use them at a time
int MIDIlock;
int LEDLock;

// MIDI messages
void noteOn(int, int);   
void noteOff(int); 

// Play Viva La Musica
void playViva(); 

void playNote(int, int);

void updateLEDs(); 

// Each of the playing cores (core 3, 4 and 5) is assigned a color
int coreColors[] = {0,0,0,COLOR_RED, COLOR_GREEN, COLOR_BLUE};

int main()                                    
{
  // The main program is core 0
  
  // The MIDI serial interface is core 1
  midi = fdserial_open(-1, MIDI_OUT_PIN, 0, 31250);
  pause(1000);
  
  // The LED interface is core 2
  leds = ws2812b_open();
   
  // Set up some locks. 
  MIDIlock = locknew();
  LEDLock = locknew();

  // Start each voice in its own core.
  // Note: cores used to be called cogs, and you still see this in Propeller code.
  // Core is now the preferred term.
  // We have already used cores 0, 1 and 2, so the next available core is core 3
  cog_run(playViva,128);      // Start core 3 to play voice 1

  pause(twoBarsRest*tempo);   // 2 bars rest

  cog_run(playViva,128);      // Start core 4 to play voice 2

  pause(twoBarsRest*tempo);   // 2 bars rest

  cog_run(playViva,128);      // Start core 5 to play voice 3  
}


// Play the round
void playViva(){
  // Play the round 3 times
  for(int j=0; j < 3; j++){
    for( int i=0; i < numNotes; i++){
      
      // Light the LED for the note.
      // Use the core ID to determine the color of the LED.
      ledsArray[vivaNotes[i]-60] = coreColors[cogid()]; 
      updateLEDs();
      
      // Play the note
      playNote(vivaNotes[i], vivaDurations[i]*tempo);
      
      // Turn off the LED for the note
      ledsArray[vivaNotes[i]-60] = COLOR_BLACK;
      updateLEDs();
    } 
  }         
}


// Play a note for a duration
void playNote(int note, int duration){
  noteOn(note, 127);
  pause(duration);
  noteOff(note);
}  

// Update the LEDs
void updateLEDs(){
  lockset(LEDLock); // Lock the LED resource
  ws2812_set(leds, LEDS_PIN, ledsArray, NUM_LEDS);
  lockclr(LEDLock); // Unlock the LED resource
}  


// MIDI messages
void noteOn(int note, int volume){
    lockset(MIDIlock);            // Lock the MIDI resource
    fdserial_txChar(midi,144);
    fdserial_txChar(midi,note);
    fdserial_txChar(midi,volume);
    lockclr(MIDIlock);           // Unlock the MIDI resource
}  

void noteOff(int note){
    lockset(MIDIlock);           // Lock the MIDI resource
    fdserial_txChar(midi,128);
    fdserial_txChar(midi,note);
    fdserial_txChar(midi,0);
    lockclr(MIDIlock);           // Unlock the MIDI resource
}  

