/*
  Jim Arlow, 2020
  Program to test MIDI output on pin 0. 
*/
#include "simpletools.h"  
#include "fdserial.h"   

#define MIDI_OUT_PIN 0  

// Viva La Musica, Micheal Praetorius (1571-1621)
// This is a three voice round.
int vivaNotes[] = {79,72,71,69,67,66,67,66,71,72,74,72,71,69,71,69,67,67,60,62,64,62};  
int vivaDurations[] = {3, 1, 2, 1, 1, 2, 2, 4, 3, 1, 2, 1, 1, 2, 2, 4, 2, 4, 2, 2, 2, 4}; // 1 = quaver, 2 = crotchet etc.  

const int numNotes = 22;

const int tempo = 200;      // The duration of a quaver in milliseconds
const int twoBarsRest = 16; // The rest between voices

// The MIDI serial output
fdserial *midi; 

// We will protect the MIDI resource with a lock
// so that only one voice can output on it at a time.
int MIDIlock;

// MIDI messages
void noteOn(int, int);   
void noteOff(int); 

// Play Viva La Musica
void playNote(int, int);
void playVivaLaMusica();  

int main()                                    
{
  midi = fdserial_open(-1, MIDI_OUT_PIN, 0, 31250);
  pause(1000);
   
  // The Propeller chip has seven available locks. 
  // lockset(int lockid) causes a core to wait until the specified lock is free and then set it. 
  // lockclr(int lockid) clears a lock so another core can use it.
  // We will use this mechanism to ensure only one cog at a time can access midi.
  MIDIlock = locknew();

  // Start each voice in its own core.
  // Note: cores used to be called cogs, and you still see this in Propeller code.
  // Core is the preferred term.
  cog_run(playVivaLaMusica,128);     // Start a core to play voice 1

  pause(twoBarsRest*tempo);          // 2 bars rest

  cog_run(playVivaLaMusica,128);     // Start a core to play voice 2

  pause(twoBarsRest*tempo);          // 2 bars rest

  cog_run(playVivaLaMusica,128);     // Start a core to play voice 3  
}


// Play the 3 part round 3 times
void playVivaLaMusica(){
  // Play the round 3 times
  for(int j=0; j < 3; j++){
    for( int i=0; i < numNotes; i++){
      playNote(vivaNotes[i], vivaDurations[i]);
    } 
  }         
}


// Play a note for a given duration
void playNote(int note, int duration){
  noteOn(note, 127);
  pause(duration*tempo);
  noteOff(note);
  
}  


// Send the MIDI Note On message
void noteOn(int note, int volume){
    lockset(MIDIlock);              // Lock the MIDI resource for this core
    fdserial_txChar(midi,144);
    fdserial_txChar(midi,note);
    fdserial_txChar(midi,volume);
    lockclr(MIDIlock);              // Unlock the MIDI resource for this core
}  

// Send the MIDI Note off message
void noteOff(int note){
    lockset(MIDIlock);              // Lock the MIDI resource for this core
    fdserial_txChar(midi,128);
    fdserial_txChar(midi,note);
    fdserial_txChar(midi,0);
    lockclr(MIDIlock);              // Unlock the MIDI resource for this core
}  